/*******************
 * game.cc - halma engine
 * Copyright (C) 2000 Christoph Groth
 *
 * This file is part of xhalma.  You are allowed to modify and/or redistribute
 * it under the terms of the GNU General Public License.  You should have
 * received a copy named COPYING along with this program, if not you can find
 * it at http://www.gnu.org/copyleft/gpl.html
 ******************/

#include <climits>
#include <vector>
#include <algorithm>
#include "defs.hh"
#include "game.hh"

void Game::do_move(const Move &m)
{
    assert((*dw)[m.from] == curr_pl);
    assert((*dw)[m.to] == NONE);
    (*dw)[m.from] = NONE;
    (*dw)[m.to] = (Player)curr_pl;
    curr_pl = (curr_pl == MAX) ? MIN : MAX;

    Score score = estimate();
    if (score.max == max_win_score)
	winnr = MAX;
    else if (score.min == min_win_score)
	winnr = MIN;
}

// Needed for sorting the scores at MIN nodes.
inline bool less(const Game::Score_move &a, const Game::Score_move &b)
{
    return (a.score.max + a.score.min) < (b.score.max + b.score.min);
}

// Needed for sorting the scores at MAX nodes.
inline bool greater(const Game::Score_move &a, const Game::Score_move &b)
{
    return (a.score.max + a.score.min) > (b.score.max + b.score.min);
}

Game::Move Game::do_ai_move(int depth, Callback callback)
{
    assert(depth > 0);

    Score score(estimate());
    std::vector<Move> *moves;
    std::vector<Score_move> *scores;
    all_moves(&moves, &scores, curr_pl, score);

    int bestc = 0, bestsc;

    int alpha = INT_MIN, beta = INT_MAX;
    if (depth > 1) {
	if (curr_pl == MAX) {
	    for (std::vector<Score_move>::iterator
		     i = scores->begin(), e = scores->end();
		 i != e; ++i) {
		Move &m = *i->move;
		(*dw)[m.from] = NONE;
		(*dw)[m.to] = MAX;
		// As minimax returns a score sum we simply set max to this
		// value and min to 0.
		if (i->score.max == max_win_score)
		    i->score.max = max_utopic + depth;
		else
		    i->score.max = minimax(MIN, i->score,
					   alpha, beta, depth-1);
		i->score.min = 0;
		set_to_max(alpha, i->score.max);
		(*dw)[m.from] = MAX;
		(*dw)[m.to] = NONE;
	    }
	} else {
	    float s = scores->size();
	    for (std::vector<Score_move>::iterator
		     i = scores->begin(), b = i, e = scores->end();
		 i != e; ++i) {
		Move &m = *i->move;
		(*dw)[m.from] = NONE;
		(*dw)[m.to] = MIN;
		if (callback && depth > 3) if (!callback(this, (i - b) / s))
		    return Move(Veci(0,0), Veci(0,0)); // abort!
//  		cout << "d1 begin (" << m.from.x << ' ' << m.from.y
//  		     << ") -> (" << m.to.x << ' ' << m.to.y
//  		     << ")   est=" << i->score.sum()
//  		     << " beta=" << beta << endl;
		// As minimax returns a score sum we simply set max to this
		// value and min to 0.
		if (i->score.min == min_win_score)
		    i->score.max = min_utopic - depth;
		else
		    i->score.max = minimax(MAX, i->score,
					   alpha, beta, depth-1);
		i->score.min = 0;
		set_to_min(beta, i->score.max);
		(*dw)[m.from] = MIN;
		(*dw)[m.to] = NONE;
//  		cout << "d1 end score=" << i->score.max
//  		     << " beta=" << beta << '\n' << endl;
	    }
	}
	// Sort again so we'll have the best moves at the beginning.
	sort(scores->begin(), scores->end(), curr_pl == MAX ? greater : less);

	// Choose one of the moves with best score.
	bestsc = scores->front().score.max;
	for (std::vector<Score_move>::const_iterator i = scores->begin();
	     i->score.max == bestsc && i != scores->end(); ++i, ++bestc) { }
    } else {
	// Choose one of the moves with best score. (depth == 1)
	bestsc = scores->front().score.sum();
	for (std::vector<Score_move>::const_iterator i = scores->begin();
	     i->score.sum() == bestsc && i != scores->end(); ++i, ++bestc) { }
    }

    cout << "Score " << bestsc << " reachable by " << bestc
	 << " of " << scores->size() << " moves." << endl;
    Move best_move = *(*scores)[rand() % bestc].move;
    do_move(best_move);

    delete moves;
    delete scores;
    return best_move;
}

Game::Game(Player beg_player)
    : curr_pl(beg_player), winnr(NONE)
{
    assert(curr_pl == MAX || curr_pl == MIN);
}

void Game::all_moves(std::vector<Move> **moves,
		     std::vector<Score_move> **scores, int player, Score score)
{
    std::vector<Move> *m = new std::vector<Move>;
    m->reserve(400);  // change!!!!!!
    std::vector<Score_move> *s = new std::vector<Score_move>;
    s->reserve(400); // ditto!!!!
    Dests dests;
    dests.reserve(40); // ditto!!!!
    Hexarr<int> &imp = (player == MAX) ? *imp_max : *imp_min;

    // Generate all moves and their scores (without searching).
    if (player == MAX) {
	for (Hexarr<Player>::iterator i = dw->begin(), e = dw->end();
	     i != e; ++i) {
	    if (*i != player) continue;
	    dests.erase(dests.begin(), dests.end());
	    // We won't need the returned Hexarr.
	    delete valid_moves(i, &dests);
	    Score score2(score.max - imp[i.get()], score.min);
	    for (Dests::iterator j = dests.begin(); j != dests.end(); ++j) {
		m->push_back(Move(i.get(), j->get()));
		Score score3(score2.max + imp[j->get()], score2.min);
		s->push_back(Score_move(score3, &m->back()));
	    }
	}
    } else {
	for (Hexarr<Player>::iterator i = dw->begin(), e = dw->end();
	     i != e; ++i) {
	    if (*i != player) continue;
	    dests.erase(dests.begin(), dests.end());
	    // We won't need the returned Hexarr.
	    delete valid_moves(i, &dests);
	    Score score2(score.max, score.min - imp[i.get()]);
	    for (Dests::iterator j = dests.begin(); j != dests.end(); ++j) {
		m->push_back(Move(i.get(), j->get()));
		Score score3(score2.max, score2.min + imp[j->get()]);
		s->push_back(Score_move(score3, &m->back()));
	    }
	}
    }

    // Now we have all moves in `moves' and their scores in `scores'.
    // We'll sort the moves by that score and this way improve Alpha-Beta
    // efficiency.
    sort(s->begin(), s->end(), player == MAX ? greater : less);

    *moves = m; *scores = s;
}

Hexarr<char> *Game::valid_moves(Veci c, Dests *m) const
{
    Hexarr<Player>::iterator dwi(dw.get(), c);
    return valid_moves(dwi, m);
}

Hexarr<char> *Game::valid_moves(const Hexarr<Player>::iterator &dwi,
				Dests *m) const
{
    assert(dwi.valid());
    Hexarr<char> *ret = dw->create_char(false);
    Hexarr<char>::iterator mi(ret, dwi.get());
    if (*dwi != NONE) {
	// Suppress moves to original position.
	*mi = true;
	recurse_move(mi, dwi, true, m);
	*mi = false;
    }
    return ret;
}

// This function calls itself recursively and changes `dw' to reflect the
// current state.  The old state is restored before return.
int Game::minimax(Player player, Score score, int alpha, int beta, int depth)
{
    assert(depth > 0);

    if (depth == 1) {
	// Don't call oneself recursively just compute the scores and prune as
	// soon as possible.
	Hexarr<int> &imp = (player == MAX) ? *imp_max : *imp_min;
	Dests dests;
	dests.reserve(40);
	if (player == MAX) {
alpha = INT_MIN;
	    for (Hexarr<Player>::iterator i = dw->begin(), e = dw->end();
		 i != e; ++i) {
		if (*i != MAX) continue;
		dests.erase(dests.begin(), dests.end());
		// We won't need the returned Hexarr.
		delete valid_moves(i, &dests);
		Score score2(score.max - imp[i.get()], score.min);
		for (Dests::iterator j = dests.begin();
		     j != dests.end(); ++j) {
		    Score score3(score2.max + imp[j->get()], score2.min);
		    if (score3.max == max_win_score)
			score3 = Score(max_utopic, 0);
		    set_to_max(alpha, score3.sum());
//  		    cout << "  d2 (" << j->get().x << ' ' << j->get().y
//  			 << ")  al=" << alpha << endl;
		    if (alpha > beta) return alpha;
		}
	    }
	    return alpha;
	} else {
beta = INT_MAX;
	    for (Hexarr<Player>::iterator i = dw->begin(), e = dw->end();
		 i != e; ++i) {
		if (*i != MIN) continue;
		dests.erase(dests.begin(), dests.end());
		// We won't need the returned Hexarr.
		delete valid_moves(i, &dests);
		Score score2(score.max, score.min - imp[i.get()]);
		for (Dests::iterator j = dests.begin();
		     j != dests.end(); ++j) {
		    Score score3(score2.max, score2.min + imp[j->get()]);
		    if (score3.min == min_win_score)
			score3 = Score(0, min_utopic);
		    set_to_min(beta, score3.sum());
		    if (alpha >= beta) return beta;
		}
	    }
	    return beta;
	}
    } else {
	std::vector<Move> *m;
	std::vector<Score_move> *s;
	all_moves(&m, &s, player, score);
	std::auto_ptr<std::vector<Move> > moves(m);
	std::auto_ptr<std::vector<Score_move> > scores(s);
	// Now we have all moves in `moves' and their scores in `scores'.
	// We'll sort the moves by that score and this way improve Alpha-Beta
	// efficiency.
	sort(scores->begin(), scores->end(), player == MAX ? greater : less);

	if (player == MAX) {
	    alpha = INT_MIN;
	    for (std::vector<Score_move>::iterator
		     i = scores->begin(), e = scores->end();
		 i != e && alpha <= beta; ++i) {
		Move &m = *i->move;
		(*dw)[m.from] = NONE;
		(*dw)[m.to] = MAX;
		if (i->score.max == max_win_score) alpha = max_utopic + depth;
		set_to_max(alpha, minimax(MIN, i->score,
					  alpha, beta, depth-1));
		(*dw)[m.from] = MAX;
		(*dw)[m.to] = NONE;
	    }
	    return alpha;
	} else {
	    beta = INT_MAX;
	    for (std::vector<Score_move>::iterator
		     i = scores->begin(), e = scores->end();
		 i != e && alpha <= beta; ++i) {
		Move &m = *i->move;
		(*dw)[m.from] = NONE;
		(*dw)[m.to] = MIN;
		if (i->score.min == min_win_score) beta = min_utopic - depth;
		set_to_min(beta, minimax(MAX, i->score, alpha, beta, depth-1));
		(*dw)[m.from] = MIN;
		(*dw)[m.to] = NONE;
	    }
	    return beta;
	}
    }
    return 0; // To suppress g++ warning.
}

Game::Score Game::estimate()
{
    Score score = Score(0, 0);
    Hexarr<Player>::iterator i = dw->begin(), end = dw->end();
    for (; i != end; ++i) {
	switch (*i) {
	case NONE:
	    break;
	case MAX:
	    score.max += (*imp_max)[i.get()];
	    break;
	case MIN:
	    score.min += (*imp_min)[i.get()];
	    break;
	}
    }
    return score;
}

// Check if there are valid moves in the direction of NEIGHBOR_FUNC.
#define EXAMINE_NEIGHBOR(NEIGHBOR_FUNC) {			\
    Hexarr<Player>::iterator dwi2 = dwi;			\
    dwi2.NEIGHBOR_FUNC();					\
    if (dwi2.valid()) {						\
	Hexarr<char>::iterator mi2 = mi;			\
	mi2.NEIGHBOR_FUNC();					\
	if (*dwi2 == NONE) {					\
	    /* Normal move: no recursion. */			\
	    if (first_move) {					\
		*mi2 = true;					\
                if (m) m->push_back(dwi2);			\
	    }							\
	} else {						\
	    /* Maybe this is a jump move. */			\
	    dwi2.NEIGHBOR_FUNC();				\
	    mi2.NEIGHBOR_FUNC();				\
	    if (dwi2.valid() && *dwi2 == NONE && !*mi2) {	\
		/* Yes, jump move: do recursion. */		\
		*mi2 = true;					\
                if (m) m->push_back(dwi2);			\
		recurse_move(mi2, dwi2, false, m);		\
	    }							\
	}							\
    }								\
}

void Game::recurse_move(const Hexarr<char>::iterator &mi,
			const Hexarr<Player>::iterator &dwi,
			bool first_move, Dests *m) const
{
    EXAMINE_NEIGHBOR(lp);
    EXAMINE_NEIGHBOR(rp);
    EXAMINE_NEIGHBOR(l);
    EXAMINE_NEIGHBOR(r);
    EXAMINE_NEIGHBOR(ln);
    EXAMINE_NEIGHBOR(rn);
}

Hex_game::Hex_game(int sidelen, int rows_per_player, Player beg_player)
    : Game(beg_player)
{
    assert(sidelen > rows_per_player);

    dw.reset(new Hex_ha<Player>(sidelen, NONE));
    ownrs.reset(new Hex_ha<Player>(sidelen, NONE));
    imp_max.reset(new Hex_ha<int>(sidelen, NONE));
    imp_min.reset(new Hex_ha<int>(sidelen, NONE));

    const int maximum_y = dw->height() - 1;
    const int target_importance = dw->height() - rows_per_player;
    for (int i=0; i < rows_per_player; ++i) {
	ownrs->set_row(i, MIN);
	ownrs->set_row(maximum_y - i, MAX);
	dw->set_row(i, MAX);
	dw->set_row(maximum_y - i, MIN);
	imp_min->set_row(i, -target_importance);
	imp_max->set_row(maximum_y - i, target_importance);
    }

    for (int i = 0, e = dw->height() - rows_per_player; i < e; ++i) {
	imp_max->set_row(i, i);
	imp_min->set_row(maximum_y - i, -i);
    }
    // = (number of fields in first and last row) * (number of rows) / 2
    max_win_score = ((2 * sidelen + rows_per_player - 1)
		     * rows_per_player / 2) * target_importance;
    min_win_score = -max_win_score;
    max_utopic = INT_MAX / 2;
    min_utopic = INT_MIN / 2;
}
