/*******************
 * game.hh - halma engine
 * Copyright (C) 2000 Christoph Groth
 *
 * This file is part of xhalma.  You are allowed to modify and/or redistribute
 * it under the terms of the GNU General Public License.  You should have
 * received a copy named COPYING along with this program, if not you can find
 * it at http://www.gnu.org/copyleft/gpl.html
 ******************/

#ifndef GAME_HH
#define GAME_HH

#include "hexarr.hh"

class Game_widget;
class Board;

enum Player { NONE, MAX, MIN};

class Game {
    friend Game_widget;
public:
    // Used for progress reporting.  If the callback functions returns false,
    // abort computations.
    typedef bool (*Callback)(const Game *, float);
    typedef std::vector<Hexarr<Player>::iterator> Dests;
    struct Move {
	Move() { }
	Move(const Veci &_from, const Veci &_to) : from(_from), to(_to) { }
	Veci from, to;
    };
    struct Score {
	Score() { }
	Score(int _max, int _min) : max(_max), min(_min) { }
	int sum() const { return max + min; }
	int max, min;
    };
    struct Score_move {
	Score_move() { }
	Score_move(Score _score, Move *_move) : score(_score), move(_move) { }
	Score score;
	Move *move;
    };

    Player current_player() const { return curr_pl; }
    Player winner() const { return winnr; }
    Move do_ai_move(int depth, Callback callback=0);
    void do_move(const Move &m); // Must be a valid move.
    // Valid moves from a given location.  Returns an Hexarr of chars with an 1
    // for every reachable field and 0 otherwise.
    Hexarr<char> *valid_moves(Veci c, Dests *m=0) const;
    Hexarr<char> *valid_moves(const Hexarr<Player>::iterator &dwi,
			      Dests *m=0) const;
protected:
    explicit Game(Player beg_player);

    // Creates a vector containing all valid moves for `player' and another
    // sorted vector (the best moves are at the beginning) containing the
    // scores and pointers to the moves.  Remember to delete both vectors!
    void all_moves(std::vector<Move> **moves, std::vector<Score_move> **scores,
		   int player, Score score);

    // Estimate situation of `player' by descending the move-graph, bigger
    // return values == better.  The return value is a combined score.
    // player = The player who is about to move.
    // score = The current heuristic score for `curr_pl' not for `player'.
    int minimax(Player player, Score score,
		int alpha, int beta, int depth);

    // Estimate the current situation by using (very simple) heuristics.  The
    // first value is the score for MAX (always positive), the second value is
    // the score for MIN (always negative).  This separation makes detecting
    // when a player wins easier.  Positive values of the sum mean MAX is
    // leading, negative mean MIN has an advantage.
    Score estimate();

    // Take a field and recursively compute all possible moves from that field.
    // mi is the iterator of that field in a Hexarr of chars denoting the
    // possible moves (1 or 0).  dwi is an iterator of the same field in the
    // Hexarr of dwellers.
    void recurse_move(const Hexarr<char>::iterator &mi,
		      const Hexarr<Player>::iterator &dwi,
		      bool first_move, Dests *m) const;

    Player curr_pl;  // current player
    Player winnr;    // winner
    std::auto_ptr<Hexarr<Player> > dw;      // dwellers
    std::auto_ptr<Hexarr<Player> > ownrs;
    std::auto_ptr<Hex_ha<int> > imp_max, imp_min; // field importance for MAX/MIN.
    // The winning scores.
    int max_win_score, min_win_score;
    // Utopic scores that can never be reached.  Used when a player is going to
    // win to ensure that he will choose this move and that his opponent will
    // avoid it.
    int max_utopic, min_utopic;
};

class Hex_game : public Game {
public:
    explicit Hex_game(int sidelen, int rows_per_player = 2,
		      Player beg_player = MAX);
};

#endif /* GAME_HH */
