/*******************
 * game_widget.cc - widget for displaying the game state
 * Copyright (C) 2000 Christoph Groth
 *
 * This file is part of xhalma.  You are allowed to modify and/or redistribute
 * it under the terms of the GNU General Public License.  You should have
 * received a copy named COPYING along with this program, if not you can find
 * it at http://www.gnu.org/copyleft/gpl.html
 ******************/

#include <qapplication.h>
#include <qpainter.h>
#include "defs.hh"
#include "game_widget.hh"

const QColor Game_widget::plcol[3] = {
    QColor(128, 128, 128),
    QColor( 64,  64, 196),
    QColor(196,  64,  64)
};

Game_widget::Game_widget(Game *game, Game::Callback _callback,
			 QWidget *parent, const char *name)
    : QWidget(parent, name), g(*game), callback(_callback), scale(0.7),
      s(g.dw->create_char(false)), getting_user_move(false), show_moves(false),
      from(-1, 0)
{
    setMinimumSize((s->width() + 1) * 20, s->height() * 17);
    setBackgroundMode(NoBackground);
}

void Game_widget::get_user_move()
{
    getting_user_move = true;
}

void Game_widget::abort_getting()
{
    s.reset(g.dw->create_char(false));
    getting_user_move = false;
    update();
}

void Game_widget::show_valid_moves(bool state)
{
    show_moves = state;
    update();
}

void Game_widget::do_ai_move(int depth)
{
    s.reset(g.dw->create_char(false));
    Game::Move m = g.do_ai_move(depth, callback);
    (*s)[m.from] = true;
    (*s)[m.to] = true;
    update();
}

void Game_widget::resizeEvent(QResizeEvent *)
{
    pmap.resize(width(), height());
    pmap.fill(backgroundColor());

    double fac = sqrt(0.75);
    d.x = (int)min((double)width() / (s->width() + 0.5),
		   height() / fac / s->height());
    d.y = int(fac * d.x);
    diam = int(scale * d.x);
    b.x = (width() - d.x * (s->width() - 1) - diam) / 2;
    b.y = (height() - d.y * (s->height() - 1) - diam) / 2;
}

void Game_widget::paintEvent(QPaintEvent *)
{
    QPainter p;
    QPen pen;
    pen.setWidth(max(diam / 8, 1));
    p.begin(&pmap);
    int y = height() - b.y - diam;
    Veci i;
    for (i.y=0; i.y < s->height(); ++i.y, y -= d.y) {
	int x = b.x - d.x / 4;
	if (i.y & 1) x += d.x / 2;
	for (i.x=0; i.x < s->width(); ++i.x, x += d.x) {
	    if (!s->valid(i)) continue;
	    int occ = (*g.dw)[i];
	    QColor col = (occ ? plcol[occ] : backgroundColor());
	    p.setBrush(col);
	    if (show_moves || from.x == -1)
		p.setBrush((*s)[i] ? col.light() : col);
	    else
		p.setBrush(i == from ? col.light() : col);
	    pen.setColor(plcol[(*g.ownrs)[i]]);
	    p.setPen(pen);
	    p.drawEllipse(x, y, diam, diam);
	}
    }
    p.end();
    bitBlt(this, 0,0, &pmap);
}

void Game_widget::mouseReleaseEvent(QMouseEvent *me)
{
    if (!getting_user_move) return;

    Veci p;
    p.y = (height() - me->y() - b.y) / d.y;
    p.x = (me->x() - b.x + (d.x - diam) / 2
	   - d.x/4 + (p.y & 1 ? 0 : d.x/2)) / d.x;
    if (!s->valid(p)) return;
    if ((*g.dw)[p] == g.curr_pl) {
	// User clicked on new from field.
	s.reset(g.valid_moves(p));
	assert(equal_dims(*s, *g.dw));
	(*s)[p] = true;
	from = p;
	update();
    } else if ((*s)[p] && from.x != -1) {
	// User clicked on a destination field.
	g.do_move(Game::Move(from, p));
	s.reset(g.dw->create_char(false));
	getting_user_move = false;
	from.x = -1;
	update();
	qApp->processEvents();
	emit user_has_moved();
    } else return; // User click is meaningless.
}
