/*******************
 * hexarr.hh - hexagonal containers
 * Copyright (C) 2000 Christoph Groth
 *
 * This file is part of xhalma.  You are allowed to modify and/or redistribute
 * it under the terms of the GNU General Public License.  You should have
 * received a copy named COPYING along with this program, if not you can find
 * it at http://www.gnu.org/copyleft/gpl.html
 ******************/

#ifndef HEXARR_HH
#define HEXARR_HH

#include <vector>
#include "vec.hh"

template <class T> class Hexarr;

// ******** Hexarr_iterator - iterator for hex arrays ********
template <class T, class P, class R>
class Hexarr_iterator {
public:
    typedef Hexarr_iterator<T, P, R> Self;

    Hexarr_iterator() : m(NULL) { }
    Hexarr_iterator(Hexarr<T> *_m, Veci _c)
	: m(_m), c(_c) { }

    friend bool operator==<T, P, R>(const Hexarr_iterator<T, P, R> &a,
				    const Hexarr_iterator<T, P, R> &b);
    R operator*() const { return (*m)[c]; }
    P operator->() const { return &(*m)[c]; }
    Self &operator++() {
	if (++c.x - m->o[c.y] < (int)m->f[c.y].size()) return *this;
	if (c.y >= m->s.y - 1) return *this; // == end()
	++c.y; c.x = m->o[c.y];
	return *this;
    }
    Self operator++(int) {
	Self ret = *this;
	++*this;
	return ret;
    }
    //   lp /\ rp
    //     /  \ 
    //     |  |
    // ^ l |  | r
    // |   \  /
    // | ln \/ rn
    // +--->
    void lp() { c.x += (c.y & 1) - 1; ++c.y; }
    void rp() { c.x += c.y & 1; ++c.y; }
    void l()  { c.x -= 1; }
    void r()  { c.x += 1; }
    void ln() { c.x += (c.y & 1) - 1; --c.y; }
    void rn() { c.x += c.y & 1; --c.y; }
    bool valid() const { return m->valid(c); }
    Veci get() const { return c; }
    Hexarr<T> mother() { return m; }
    const Hexarr<T> mother() const { return m; }
private:
    Hexarr<T> *m;
    Veci c;
};

template <class T, class P, class R>
bool operator==(const Hexarr_iterator<T, P, R> &a,
		const Hexarr_iterator<T, P, R> &b)
{
    return a.m == b.m && a.c == b.c;
}

// ******** Hexarr - base class for hex arrays ********
template <class T>
class Hexarr {
    template <class T1, class T2>
    friend bool equal_dims(const Hexarr<T1> &a, const Hexarr<T2> &b);
public:
    typedef T value_type;
    typedef Hexarr_iterator<T, T*, T&> iterator;
    typedef Hexarr_iterator<T, const T*, const T&> const_iterator;
    friend iterator;
    friend const_iterator;
    virtual Hexarr<T> *clone() const = 0;
    virtual Hexarr<T> *create(const T &elem_init) const = 0;
    virtual Hexarr<char> *create_char(char elem_init) const = 0;
    virtual ~Hexarr() { }
    bool valid(Veci c) const;
    int width() const { return s.x; }
    int height() const { return s.y; }
    Veci size() const { return s; }
    iterator begin() { return t_begin<iterator>(); }
    iterator end() { return t_end<iterator>(); }
    const_iterator begin() const { return t_begin<const_iterator>(); }
    const_iterator end() const { return t_end<const_iterator>(); }
    T &operator[](Veci c) { return f[c.y][c.x - o[c.y]]; }
    const T &operator[](Veci c) const { return f[c.y][c.x - o[c.y]]; }
    void set_row(int row, const T &val) {
	std::fill(f[row].begin(), f[row].end(), val);
    }
protected:
    // Only to be used by constructors of derived classes.
    Hexarr(Veci size) : s(size), o(s.y), f(s.y) { }
    Veci s;
    // 0 1 1 1 1 1 1
    //  0 1 1 1 1 1    This line has a offset of 1.
    // 0 0 1 1 1 1 1   The first line is displaced to the left.
    std::vector<int> o;              // the offsets
    std::vector<std::vector<T> > f;  // the fields
private:
    template <class I> I t_begin() { return I(this, Veci(o[0], 0)); }
    template <class I> I t_end() {
	return I(this, Veci(o[s.y-1] + f[s.y-1].size(), s.y-1));
    }
};

template <class T>
bool Hexarr<T>::valid(Veci c) const
{
    if (c.y < 0 || c.y >= (int)o.size()) return false;
    int i = c.x - o[c.y];
    if (i < 0 || i >= (int)f[c.y].size()) return false;
    return true;
}

template <class T1, class T2>
bool equal_dims(const Hexarr<T1> &a, const Hexarr<T2> &b)
{
    if (a.s != b.s) return false;
    if (a.o != b.o) return false;
    std::vector<std::vector<T1> >::const_iterator ia = a.f.begin(), ea;
    std::vector<std::vector<T2> >::const_iterator ib = b.f.begin();
    for (ea = a.f.end();  ia != ea; ++ia, ++ib)
	if (ia->size() != ib->size()) return false;
    return true;
}

// ******** Hex_rect - a rectagonally shaped hex array ********
//  ____
// |    |
// |    |
// |    |
// |____|
//
template <class T>
class Rect_ha : public Hexarr<T> {
public:
    Rect_ha(Veci size, const T &elem_init);
    Rect_ha<T> *clone() const { return new Rect_ha<T>(*this); }
    Rect_ha<T> *create(const T &elem_init) const {
	return new Rect_ha<T>(s, elem_init);
    }
    Rect_ha<char> *create_char(char elem_init) const {
	return new Rect_ha<char>(s, elem_init);
    }
};

template <class T>
Rect_ha<T>::Rect_ha(Veci size, const T &elem_init)
    : Hexarr<T>(size)
{
    std::fill(f.begin(), f.end(), std::vector<T>(w, elem_init));
}

// ******** Hex_ha - a hexagonally shaped hex array ********
//   ___
//  /   \ 
// /     \ 
// \     /
//  \___/
//
template <class T>
class Hex_ha : public Hexarr<T> {
public:
    Hex_ha(int sidel, const T &elem_init);
    Hex_ha<T> *clone() const { return new Hex_ha<T>(*this); }
    Hex_ha<T> *create(const T &elem_init) const {
	return new Hex_ha<T>(sl, elem_init);
    }
    Hex_ha<char> *create_char(char elem_init) const {
	return new Hex_ha<char>(sl, elem_init);
    }
protected:
    int sl;
};

template <class T>
Hex_ha<T>::Hex_ha(int sidel, const T &elem_init)
    : Hexarr<T>(Veci(2*sidel - 1, 2*sidel - 1)), sl(sidel)
{
    assert(sl > 1);

    int i = sl - 1;
    int offs = 0;
    o[i] = offs; f[i].insert(f[i].end(), sl + i, elem_init);  // sl + i == w
    int j = i-- + 1;
    for (; i >= 0; --i, ++j) {
	offs += !(i&1);
	o[i] = offs; f[i].insert(f[i].end(), sl + i, elem_init);
	o[j] = offs; f[j].insert(f[j].end(), sl + i, elem_init);
    }
}

// ******** Hex_star - a six-beam-star-shaped hex array ********
//        p
//   lp __/\__ rp
// ^    \    /
// | ln /_  _\ rn
// |      \/
// +--->  n
//
template <class T>
class Star_ha : public Hexarr<T> {
public:
    Star_ha(int beam_sidel, const T &elem_init);
    Star_ha<T> *clone() const { return new Star_ha<T>(*this); }
    Star_ha<T> *create(const T &elem_init) const {
	return new Star_ha<T>(bsl, elem_init);
    }
    Star_ha<char> *create_char(char elem_init) const {
	return new Star_ha<char>(bsl, elem_init);
    }
    void fill_p(const T &val);
    void fill_n(const T &val);
    void fill_lp(const T &val);
    void fill_rp(const T &val);
    void fill_ln(const T &val);
    void fill_rn(const T &val);
protected:
    int bsl;
};
#endif /* HEXARR_HH */
