/*******************
 * main_wnd.cc - main window widget
 * Copyright (C) 2000 Christoph Groth
 *
 * This file is part of xhalma.  You are allowed to modify and/or redistribute
 * it under the terms of the GNU General Public License.  You should have
 * received a copy named COPYING along with this program, if not you can find
 * it at http://www.gnu.org/copyleft/gpl.html
 ******************/

#include <qapplication.h>
#include <qpushbutton.h>
#include <qcombobox.h>
#include <qlabel.h>
#include <qcheckbox.h>
#include <qprogressbar.h>
#include <qmessagebox.h>
#include "defs.hh"
#include "main_wnd.hh"
#include "game_widget.hh"

const int Main_wnd::pbar_stepc = 1024;
std::map<const Game *, Main_wnd *> Main_wnd::windows;

Main_wnd::Main_wnd(QWidget *parent = NULL, const char *name = NULL)
    : QWidget(parent, name), gw(0), buttons_w(145), min_buttons_h(320),
      cont_comp(true)
{
    ai_levels.push_back(Ai_level("human", -1));
    ai_levels.push_back(Ai_level("AI - silly (1)", 1));
    ai_levels.push_back(Ai_level("AI - foolish (2)", 2));
    ai_levels.push_back(Ai_level("AI - fair (3)", 3));
    ai_levels.push_back(Ai_level("AI - good (4)", 4));
    ai_levels.push_back(Ai_level("AI - smart (5)", 5));
    ai_levels.push_back(Ai_level("AI - challenging (6)", 6));

    red_label = new QLabel("red player:", this);
    red_combo = new QComboBox(this, "red combo");
    blue_label = new QLabel("blue player:", this);
    blue_combo = new QComboBox(this, "blue combo");
    blue_combo->setEnabled(false);  // Doesn't work yet.
    for (std::vector<Ai_level>::const_iterator i = ai_levels.begin();
	 i != ai_levels.end(); ++i) {
	blue_combo->insertItem(i->first.c_str());
	red_combo->insertItem(i->first.c_str());
    }
    red_combo->setCurrentItem(4);
    pbar = new QProgressBar(pbar_stepc, this, "pbar");
    show_moves = new QCheckBox("show valid moves", this, "show_moves");

    cont = new QPushButton("continue", this, "continue");
    connect(cont, SIGNAL(clicked()), this, SLOT(user_move_done()));
    cont->hide();  // Doesn't work yet, too.
    restart = new QPushButton("restart", this, "restart");
    connect(restart, SIGNAL(clicked()), this, SLOT(do_restart()));
    rules = new QPushButton("help on rules", this, "rules");
    connect(rules, SIGNAL(clicked()), this, SLOT(show_rules()));
    about = new QPushButton("about", this, "about");
    connect(about, SIGNAL(clicked()), this, SLOT(show_about()));
    quit = new QPushButton("quit", this, "quit");
    connect(quit, SIGNAL(clicked()), this, SLOT(dont_cont_comp()));
    connect(quit, SIGNAL(clicked()), qApp, SLOT(quit()));

    do_restart();
}

Main_wnd::~Main_wnd()
{
    windows.erase(windows.find(g.get()));
}

void Main_wnd::user_move_done()
{
    if (is_victory()) return;
    int curr_item = (g->current_player() == MAX)
	? blue_combo->currentItem() : red_combo->currentItem();
    int other_item = (g->current_player() == MIN)
	? blue_combo->currentItem() : red_combo->currentItem();

    int depth;
    if ((depth = ai_levels[curr_item].second) == -1) {
	// human
	gw->get_user_move();
	return;
    } else {
	// AI
	set_enabled(false);
	gw->do_ai_move(depth);
	pbar->reset();
	set_enabled(true);
	if (is_victory()) return;
	// Get human move if AI plays against human.  get_user_move will call
	// user_move_done when finished.  If AI plays against AI we'll wait
	// until the players presses the continue button and
	if (ai_levels[other_item].second == -1) {
	    gw->get_user_move();
	    is_victory();
	}
    }
}

void Main_wnd::do_restart()
{
    delete gw;
    if (g.get()) windows.erase(windows.find(g.get()));
    g.reset(new Hex_game(5, 2));
    windows[g.get()] = this;
    gw = new Game_widget(g.get(), Main_wnd::set_progress, this, "game widget");
    gw->show_valid_moves(show_moves->isChecked());
    setMinimumSize(gw->minimumWidth() + buttons_w + 30,
		   max(min_buttons_h, gw->minimumHeight()) + 20);
    connect(gw, SIGNAL(user_has_moved()), this, SLOT(user_move_done()));
    connect(show_moves, SIGNAL(toggled(bool)),
	    gw, SLOT(show_valid_moves(bool)));
    gw->show();
    resizeEvent(0);
    user_move_done();
}

#ifndef VERSION
#define VERSION "(unknown version)"
#endif

const char *const Main_wnd::silly_text =
"xhalma "VERSION" - a quick and dirty halma game\n\n"
"Copyright (C) 2000 Christoph Groth <cgroth@gmx.net>\n"
"This is free software, see the file COPYING for details.\n";

void Main_wnd::show_about()
{
    QMessageBox::information(this, "xhalma - about", silly_text);
}

const char *const Main_wnd::rules_text =
"The object of Halma is to move one's pieces to the destination fields\n"
"before the opponent can achieve the same for his pieces.  The destination\n"
"fields and the pieces which have to get there share the same color.  At the\n"
"beginning one's pieces are located at the opponent's destination fields.\n\n"
"Pieces can be moved in two ways:\n"
"- Sliding to a free neighbor field.\n"
"- Jumping over other pieces.  You can jump as often as you can in one\n"
"move if there is always another piece to jump over.\n\n"
"Clearly jumping is faster and you should try to jump as far as possible\n"
"while at the same time blocking the opponent's jumps.\n\n"
"That's all.  Good luck!";

void Main_wnd::show_rules()
{
    QMessageBox::information(this, "xhalma - game rules", rules_text);
}

bool Main_wnd::set_progress(const Game *game, float progress)
{
    Main_wnd *wnd = windows[game];
    wnd->pbar->setProgress(static_cast<int>(progress * pbar_stepc));
    qApp->processEvents();
    return wnd->cont_comp;
}

// Don't forget to adjust buttons_w and min_buttons_h in the Main_wnd ctor
// after adding/removing widgets.
void Main_wnd::resizeEvent(QResizeEvent *)
{
    const int b = 10;   // outer border
    const int lh = 15;  // height of a label
    const int bh = 26;  // height of a button
    const int vd = 6;   // vertical distance between widgets

    gw->setGeometry(buttons_w + 2*b, 10,
		    width() - buttons_w - 3*b, height() - 2*b);

    int y = b;
    red_label->setGeometry(b, y, buttons_w, lh);
    y += lh + vd;
    red_combo->setGeometry(b, y, buttons_w, bh);
    y += bh + lh;
    blue_label->setGeometry(b, y, buttons_w, lh);
    y += lh + vd;
    blue_combo->setGeometry(b, y, buttons_w, bh);
    y += bh + lh;
    pbar->setGeometry(b, y, buttons_w, bh);
    y += bh + vd;
    show_moves->setGeometry(b, y, buttons_w, lh);
    y += lh + vd;
    cont->setGeometry(b, y, buttons_w, bh);

    y = height() - b - bh;
    quit->setGeometry(b, y, buttons_w, bh);
    y -= bh + vd;
    about->setGeometry(b, y, buttons_w, bh);
    y -= bh + vd;
    rules->setGeometry(b, y, buttons_w, bh);
    y -= bh + vd;
    restart->setGeometry(b, y, buttons_w, bh);
}

void Main_wnd::dont_cont_comp()
{
    cont_comp = false;
}

void Main_wnd::set_enabled(bool enable)
{
    cont->setEnabled(enable);
    restart->setEnabled(enable);
//      blue_combo->setEnabled(enable);
    red_combo->setEnabled(enable);
    show_moves->setEnabled(enable);
}

bool Main_wnd::is_victory()
{
    if (Player w = g->winner()) {
	    bool b_is_ai = ai_levels[blue_combo->currentItem()].second != -1;
	    bool r_is_ai = ai_levels[red_combo->currentItem()].second != -1;
	    bool both_human = !b_is_ai && !r_is_ai;
	    const char *t;
	if (w == MAX) {
	    if (b_is_ai)
		t = "The blue player has won!\nBetter luck next time!";
	    else if (both_human)
		t = "Congratulations!\nThe blue are victorious!";
	    else
		t = "Congratulations!\nYou are victorious!";
	    QMessageBox::information(this, "xhalma - blue wins", t);
	} else {
	    if (r_is_ai)
		t = "The red player has won!\nBetter luck next time!";
	    else if (both_human)
		t = "Congratulations!\nThe red are victorious!";
	    else
		t = "Congratulations!\nYou are victorious!";
	    QMessageBox::information(this, "xhalma - red wins", t);
	}
	do_restart();
	return true;
    } else {
	return false;
    }
}
